@@ -79,6 +79,7 @@ class PatchSet(object):
         self.name           = name
         self.variable       = None
         self.directory      = directory
+        self.config         = []
         self.fixes          = []
         self.changes        = []
         self.disabled       = False
@@ -104,6 +105,16 @@ def _unique(iterable, key=None):
     # _unique('ABBCcAD', str.lower) --> A B C A D
     return itertools.imap(next, itertools.imap(operator.itemgetter(1), itertools.groupby(iterable, key)))
 
+def _binomial(n, k):
+    "Compute binomial coefficient."
+    num = 1
+    div = 1
+    for t in xrange(1, min(k, n - k) + 1):
+        num *= n
+        div *= t
+        n -= 1
+    return num // div
+
 def _split_seq(iterable, size):
     """Split an iterator into chunks of a given size."""
     it = iter(iterable)
@@ -178,12 +189,27 @@ def load_patchsets():
     for name, directory in enum_patchsets(config.path_patches):
         patch = PatchSet(name, directory)
 
+        # Load the definition file
+        try:
+            with open(os.path.join(directory, "definition")) as fp:
+                for line in fp:
+                    if line.startswith("#"):
+                        continue
+                    tmp = line.split(":", 1)
+                    if len(tmp) != 2:
+                        continue
+                    patch.config.append((tmp[0].lower(), tmp[1].strip()))
+        except IOError:
+            pass
+
         # Enumerate .patch files in the given directory, enumerate individual patches and affected files
         for f in sorted(os.listdir(directory)):
             if not re.match("^[0-9]{4}-.*\\.patch$", f):
                 continue
             if f.startswith(config.path_IfDefined):
                 continue
+            if ("exclude", f) in patch.config:
+                continue
             if not os.path.isfile(os.path.join(directory, f)):
                 continue
             patch.files.append(f)
@@ -202,24 +228,10 @@ def load_patchsets():
 
     # Now read the definition files in a second step
     for i, patch in all_patches.iteritems():
-        filename = os.path.join(os.path.join(config.path_patches, patch.name), "definition")
-        try:
-            with open(filename) as fp:
-                content = fp.read()
-        except IOError:
-            continue # Skip this definition file
-
-        for line in content.split("\n"):
-            if line.startswith("#"):
-                continue
-            tmp = line.split(":", 1)
-            if len(tmp) != 2:
-                continue
-
-            key, val = tmp[0].lower(), tmp[1].strip()
+        for key, val in patch.config:
             if key == "depends":
                 if not name_to_id.has_key(val):
-                    raise PatchUpdaterError("Definition file %s references unknown dependency %s" % (filename, val))
+                    raise PatchUpdaterError("Definition file for %s references unknown dependency %s" % (patch.name, val))
                 patch.depends.add(name_to_id[val])
 
             elif key == "apply-after":
@@ -244,11 +256,14 @@ def load_patchsets():
             elif key == "disabled":
                 patch.disabled = _parse_int(val)
 
+            elif key == "exclude":
+                pass # Already processed above
+
             elif key == "ifdefined":
                 patch.ifdefined = val
 
             else:
-                print "WARNING: Ignoring unknown command in definition file %s: %s" % (filename, line)
+                print "WARNING: Ignoring unknown command in definition file for %s: %s" % (patch.name, line)
 
     # Filter autodepends on disabled patchsets
     for i, patch in all_patches.iteritems():
@@ -431,11 +446,11 @@ def check_bug_status(all_patches, sync_bugs=False):
 
 def generate_ifdefined(all_patches, skip_checks=False):
     """Update autogenerated ifdefined patches, which can be used to selectively disable features at compile time."""
-    enabled_patches = dict([(i, patch) for i, patch in all_patches.iteritems() if not patch.disabled])
-
-    for i, patch in enabled_patches.iteritems():
+    for i, patch in all_patches.iteritems():
         if patch.ifdefined is None:
             continue
+        if patch.disabled:
+            continue
 
         filename = os.path.join(patch.directory, config.path_IfDefined)
         headers = { 'author': "Wine Staging Team",
@@ -455,12 +470,12 @@ def generate_ifdefined(all_patches, skip_checks=False):
                 fp.write("    %s <%s>\n" % (author, email))
             fp.write("\n")
 
-            depends = resolve_dependencies(enabled_patches, i)
+            depends = resolve_dependencies(all_patches, i)
             for f in sorted(patch.modified_files):
 
                 # Reconstruct the state after applying the dependencies
                 original = get_wine_file(f)
-                selected_patches = select_patches(enabled_patches, depends, f)
+                selected_patches = select_patches(all_patches, depends, f)
                 failed = []
 
                 try:
@@ -571,9 +586,18 @@ def generate_apply_order(all_patches, skip_checks=False):
                     dependency_cache[filename].remove(unique_hash)
                     continue
 
-            # Show a progress bar while applying the patches - this task might take some time
             chunk_size = 20
-            with progressbar.ProgressBar(desc=filename, total=2 ** len(indices) / chunk_size) as progress:
+            iterables = []
+            total = 0
+            for i in xrange(1, len(indices) + 1):
+                # HACK: It is no longer feasible to check all combinations for configure.ac.
+                # Only check corner cases (applying individual patches and applying all patches).
+                if filename == "configure.ac" and i > 4 and i <= len(indices) - 4: continue
+                iterables.append(itertools.combinations(indices, i))
+                total += _binomial(len(indices), i)
+
+            # Show a progress bar while applying the patches - this task might take some time
+            with progressbar.ProgressBar(desc=filename, total=total / chunk_size) as progress:
 
                 def test_apply(current):
                     set_apply = [(i, all_patches[i]) for i in current]
@@ -601,9 +625,6 @@ def generate_apply_order(all_patches, skip_checks=False):
                             return current
                     return None
 
-                iterables = []
-                for i in xrange(0, len(indices) + 1):
-                    iterables.append(itertools.combinations(indices, i))
                 it = _split_seq(itertools.chain(*iterables), chunk_size)
                 for k, failed in enumerate(pool.imap_unordered(test_apply_seq, it)):
                     if failed is not None:
